"""
Modified version of the google analytics to s3 operator
Source of original: https://github.com/airflow-plugins/google_analytics_plugin/blob/master/operators/google_analytics_reporting_to_s3_operator.py
Accessed 5 March 2019
"""
from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC

from ewah.hooks.base import EWAHBaseHook as BaseHook

import json
import time
from datetime import timedelta
from apiclient.discovery import build
from oauth2client.service_account import ServiceAccountCredentials as SAC


class EWAHGAOperator(EWAHBaseOperator):

    _NAMES = ["ga", "google_analytics"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: False,
        EC.ES_INCREMENTAL: True,
    }

    _API_CORE_V3 = "core_v3"
    _API_CORE_V4 = "core_v4"
    _API_MULTI = "multichannel"

    _ACCEPTED_API = [
        # _API_CORE_V3,
        _API_CORE_V4,
        # _API_MULTI,
    ]

    _SAMPLE_JSON = {
        "client_secrets": {
            "type": "service_account",
            "project_id": "abc-123",
            "private_key_id": "123456abcder",
            "private_key": "-----BEGIN PRIVATE KEY-----\nxxx\n-----END PRIVATE KEY-----\n",
            "client_email": "xyz@abc-123.iam.gserviceaccount.com",
            "client_id": "123457",
            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
            "token_uri": "https://oauth2.googleapis.com/token",
            "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
            "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/xyz%40abc-123.iam.gserviceaccount.com",
        }
    }

    def __init__(
        self,
        api,  # one of _API_CORE_V3, _API_CORE_V4, _API_MULTI
        view_id,
        dimensions,
        metrics,
        page_size=10000,
        include_empty_rows=True,
        sampling_level=None,
        *args,
        **kwargs
    ):
        if kwargs.get("update_on_columns"):
            raise Exception(
                "update_on_columns supplied, but the field is "
                + "auto-generated by the operator!"
            )
        if not api in self._ACCEPTED_API:
            raise Exception(
                "api must be one of these: {0}".format(
                    ", ".join(self._ACCEPTED_API),
                )
            )

        if api == self._API_MULTI:
            shorthand = "mcf:"
        else:
            shorthand = "ga:"

        dimensions = [
            ("" if dim.startswith(shorthand) else shorthand) + dim for dim in dimensions
        ]
        metrics = [
            ("" if metric.startswith(shorthand) else shorthand) + metric
            for metric in metrics
        ]

        kwargs.update({"update_on_columns": [dim[3:] for dim in dimensions]})

        self.api = api
        self.view_id = view_id
        self.sampling_level = sampling_level
        self.dimensions = dimensions
        self.metrics = metrics
        self.page_size = page_size
        self.include_empty_rows = include_empty_rows

        self.metricMap = {
            "METRIC_TYPE_UNSPECIFIED": "varchar(255)",
            "CURRENCY": "decimal(20,5)",
            "INTEGER": "int(11)",
            "FLOAT": "decimal(20,5)",
            "PERCENT": "decimal(20,5)",
            "TIME": "time",
        }

        super().__init__(*args, **kwargs)

        credentials = BaseHook.get_connection(self.source_conn_id).extra_dejson

        if not credentials.get("client_secrets"):
            _msg = "Google Analytics Credentials misspecified!"
            _msg += " Example of a correct specifidation: {0}".format(
                json.dumps(self._SAMPLE_JSON)
            )
            for key in self._SAMPLE_JSON["client_secrets"]:
                if not key in credentials:
                    raise Exception(_msg)

        if len(dimensions) > 7:
            raise Exception(
                (
                    "Can only fetch up to 7 dimensions!" + " Currently {0} Dimensions"
                ).format(
                    str(len(dimensions)),
                )
            )

        if len(metrics) > 10:
            raise Exception(
                (
                    "Can only fetch up to 10 metrics!" + " Currently {0} Dimensions"
                ).format(
                    str(len(metrics)),
                )
            )

        if self.page_size > 10000:
            raise Exception("Please specify a page size equal to or lower than 10000.")

    def ewah_execute(self, context):
        credentials = self.source_conn.extra_dejson
        credentials = credentials.get("client_secrets", credentials)
        self.log.info("Connecting to Google...")
        if self.api == self._API_CORE_V3:
            raise Exception("Not yet implemented")
        elif self.api == self._API_CORE_V4:
            kwargs = {
                "service_object": build(
                    "analyticsreporting",
                    "v4",
                    credentials=SAC.from_json_keyfile_dict(
                        credentials,
                        ["https://www.googleapis.com/auth/analytics.readonly"],
                    ),
                ),
                "context": context,
            }
            data_func = self.get_data_v4
        elif self.api == self._API_MULTI:
            raise Exception("Not yet implemented!")
            # Open: how to parse raw data to uploadable format? See get_data_multi
            kwargs = {
                "service_object": build(
                    "analytics",
                    "v3",
                    credentials=SAC.from_json_keyfile_dict(
                        credentials,
                        ["https://www.googleapis.com/auth/analytics.readonly"],
                    ),
                ),
                "context": context,
            }
            data_func = self.get_data_multi
        else:
            raise Exception("Not yet implemented!")

        self.since_formatted = self.data_from.strftime("%Y-%m-%d")
        self.until_formatted = self.data_until.strftime("%Y-%m-%d")
        self.log.info(
            "Loading data from {0} through {1}".format(
                self.since_formatted,
                self.until_formatted,
            )
        )

        self.upload_data(data_func(**kwargs))

    def get_data_multi(self, service_object, context):
        start_index = 1
        payload = {
            "ids": "ga:" + self.view_id,
            "start_date": self.since_formatted,
            "end_date": self.until_formatted,
            "metrics": ",".join(self.metrics),
            "dimensions": ",".join(self.dimensions),
            "sort": ",".join(self.dimensions),
            # 'filters': not implemented!
            # 'samplingLevel': Default!
            "start_index": start_index,
            "max_results": self.page_size,
        }

        self.log.info("Getting data from page 1...")
        i = 1
        response = service_object.data().mcf().get(**payload).execute()
        data = response.get("rows", [])
        while response.get("nextLink"):
            i += 1
            self.log.info("Getting data from page {0}...".format(str(i)))
            start_index += self.page_size
            payload.update({"start_index": start_index})
            response = service_object.data().mcf().get(**payload).execute()
            data += response, get("rows", [])

        uploadable_data = []  # TBU

        return uploadable_data

    def get_data_v4(self, service_object, context):
        reportRequest = {
            "viewId": self.view_id,
            "dateRanges": [
                {
                    "startDate": self.since_formatted,
                    "endDate": self.until_formatted,
                }
            ],
            "samplingLevel": self.sampling_level,
            "dimensions": [{"name": d} for d in self.dimensions],
            "metrics": [{"expression": m} for m in self.metrics],
            "pageSize": self.page_size,
            "includeEmptyRows": self.include_empty_rows,
        }

        response = (
            service_object.reports()
            .batchGet(body={"reportRequests": [reportRequest]})
            .execute()
        )

        if response.get("reports"):
            report = response["reports"][0]
            rows = report.get("data", {}).get("rows", [])

            while report.get("nextPageToken"):
                time.sleep(1)
                reportRequest.update({"pageToken": report["nextPageToken"]})
                response = (
                    service_object.reports()
                    .batchGet(body={"reportRequests": [reportRequest]})
                    .execute()
                )
                report = response["reports"][0]
                rows.extend(report.get("data", {}).get("rows", []))

            if report["data"]:
                report["data"]["rows"] = rows
        else:
            report = {}

        columnHeader = report.get("columnHeader", {})
        # Right now all dimensions are hardcoded to varchar(255), will need a
        # map if any non-varchar dimensions are used in the future
        # Unfortunately the API does not send back types for Dimensions like it
        # does for Metrics (yet..)
        dimensionHeaders = [
            {
                "name": header.replace("ga:", ""),
                "type": "varchar(255)",
            }
            for header in columnHeader.get("dimensions", [])
        ]
        metricHeaders = [
            {
                "name": entry.get("name").replace("ga:", ""),
                "type": self.metricMap.get(entry.get("type"), "varchar(255)"),
            }
            for entry in columnHeader.get("metricHeader", {}).get(
                "metricHeaderEntries", []
            )
        ]

        uploadable_data = []
        rows = report.get("data", {}).get("rows", [])
        for row_counter, row in enumerate(rows):
            root_data_obj = {}
            dimensions = row.get("dimensions", [])
            metrics = row.get("metrics", [])

            for index, dimension in enumerate(dimensions):
                header = dimensionHeaders[index].get("name")  # .lower()
                root_data_obj[header] = dimension

            for metric in metrics:
                data = {}
                data.update(root_data_obj)

                for index, value in enumerate(metric.get("values", [])):
                    header = metricHeaders[index].get("name")  # .lower()
                    data[header] = value

                data["view_id"] = self.view_id

                uploadable_data += [data]

        return uploadable_data



from ewah.constants import EWAHConstants as EC
from ewah.operators.base import EWAHBaseOperator


class EWAHMongoDBOperator(EWAHBaseOperator):

    _NAMES = ["mongo", "mongodb"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
    }

    _REQUIRES_COLUMNS_DEFINITION = False

    def __init__(
        self,
        source_collection_name=None,  # defaults to target_table_name
        source_database_name=None,  # Not required if specified in connection
        timestamp_field=None,  # required for use with data_from and data_until
        single_column_mode=False,  # If True, throw all data as json into one col
        batch_size=100000,
        *args,
        **kwargs
    ):

        self.source_collection_name = (
            source_collection_name or kwargs["target_table_name"]
        )
        self.source_database_name = source_database_name

        if single_column_mode:
            if kwargs.get("columns_definition"):
                raise Exception(
                    "single_column_mode is not compatible with " + "columns_definition!"
                )
            if not kwargs.get("extract_strategy") == EC.ES_FULL_REFRESH:
                raise Exception(
                    "single_column_mode is only compatible with "
                    + "extract_strategy = {0}!".format(EC.ES_FULL_REFRESH)
                )
            if kwargs.get("update_on_columns"):
                raise Exception(
                    "single_column_mode is not compatible with " + "update_on_columns!"
                )
            if kwargs.get("primary_key_column_name"):
                raise Exception(
                    "single_column_mode is not compatible with "
                    + "primary_key_column_name!"
                )
        self.single_column_mode = single_column_mode

        super().__init__(*args, **kwargs)

        if (
            self.load_data_from
            or self.load_data_until
            or self.reload_data_from
            or (self.extract_strategy == EC.ES_INCREMENTAL)
        ):
            if not timestamp_field:
                err_msg = "If using reload_data_from, load_data_from, or load_"
                err_msg += "data_until, you must also specify timestamp_field!"
                raise Exception(err_msg)
        self.timestamp_field = timestamp_field
        self.batch_size = batch_size

    def upload_data(self, data):
        if self.single_column_mode:
            # load all data into a single column called 'document'
            data = [{"document": x} for x in data]
            super().upload_data(data)
        else:
            # normal data loading mode
            super().upload_data(data)

    def ewah_execute(self, context):
        base_filters = []
        # data_from and data_until filter expression
        if self.timestamp_field:
            if self.data_from:
                base_filters += [{self.timestamp_field: {"$gte": self.data_from}}]
            if self.data_until:
                base_filters += [{self.timestamp_field: {"$lt": self.data_until}}]
        if base_filters:
            filter_expressions = {"$and": base_filters}
        else:
            filter_expressions = None

        for batch in self.source_hook.get_data_in_batches(
            collection=self.source_collection_name,
            database=self.source_database_name,
            filter_expression=filter_expressions,
            batch_size=self.batch_size,
        ):
            self.upload_data(batch)



from airflow import DAG

from ewah.constants import EWAHConstants as EC
from ewah.ewah_utils.airflow_utils import etl_schema_tasks, datetime_utcnow_with_tz
from ewah.operators.base import EWAHBaseOperator

from collections.abc import Iterable
from copy import deepcopy
from datetime import datetime, timedelta
from typing import Optional, Type, Callable, List, Tuple, Union

import re


def dag_factory_atomic(
    dag_name: str,
    dwh_engine: str,
    dwh_conn_id: str,
    start_date: datetime,
    el_operator: Type[EWAHBaseOperator],
    operator_config: dict,
    target_schema_name: str,
    target_schema_suffix: str = "_next",
    target_database_name: Optional[str] = None,
    default_args: Optional[dict] = None,
    schedule_interval: timedelta = timedelta(days=1),
    end_date: Optional[datetime] = None,
    read_right_users: Optional[Union[List[str], str]] = None,
    additional_dag_args: Optional[dict] = None,
    additional_task_args: Optional[dict] = None,
    logging_func: Optional[Callable] = None,
    **kwargs
) -> Tuple[DAG]:
    def raise_exception(msg: str) -> None:
        """Add information to error message before raising."""
        raise Exception("DAG: {0} - Error: {1}".format(dag_name, msg))

    logging_func = logging_func or print

    if kwargs:
        logging_func("unused config: {0}".format(str(kwargs)))

    additional_dag_args = additional_dag_args or {}
    additional_task_args = additional_task_args or {}

    if not read_right_users is None:
        if isinstance(read_right_users, str):
            read_right_users = [u.strip() for u in read_right_users.split(",")]
        if not isinstance(read_right_users, Iterable):
            raise_exception("read_right_users must be an iterable or string!")

    # fake catchup = True: between start_date and end_date is only one schedule_interval
    # --> run the full refreshs every schedule_interval at the same time instead of
    # having a drift in execution time!
    if end_date:
        end_date = min(end_date, datetime_utcnow_with_tz())
    else:
        end_date = datetime_utcnow_with_tz()
    start_date += int((end_date - start_date) / schedule_interval) * schedule_interval
    if start_date == end_date:
        # if the division result is a precise integer, that implies a definite end_date
        # --> adjust to get exactly one schedule_interval delta between start_date and
        # end_date to have one last run available (that should have run before end_date)
        start_date -= schedule_interval
    else:
        # Airflow executes at the END of the execution_date - start_date has to be
        # between exactly 1 and below 2 time schedule_interval before end_date!
        # end_date - 2*schedule_interval < start_date <= end_date - schedule_interval
        start_date -= schedule_interval
        # Make sure only one execution every runs scheduled but manual triggers work!
        end_date = start_date + 2 * schedule_interval - timedelta(seconds=1)

    dag = DAG(
        dag_name,
        catchup=True,  # See above
        default_args=default_args,
        max_active_runs=1,
        schedule_interval=schedule_interval,
        start_date=start_date,
        end_date=end_date,
        **additional_dag_args,
    )

    kickoff, final = etl_schema_tasks(
        dag=dag,
        dwh_engine=dwh_engine,
        dwh_conn_id=dwh_conn_id,
        target_schema_name=target_schema_name,
        target_schema_suffix=target_schema_suffix,
        target_database_name=target_database_name,
        read_right_users=read_right_users,
        **additional_task_args,
    )

    base_config = deepcopy(additional_task_args)
    base_config.update(operator_config.get("general_config", {}))
    with dag:
        for table in operator_config["tables"].keys():
            table_config = deepcopy(base_config)
            table_config.update(operator_config["tables"][table] or {})
            table_config.update(
                {
                    "task_id": "extract_load_" + re.sub(r"[^a-zA-Z0-9_]", "", table),
                    "dwh_engine": dwh_engine,
                    "dwh_conn_id": dwh_conn_id,
                    "extract_strategy": table_config.get(  # Default to full refresh
                        "extract_strategy", EC.ES_FULL_REFRESH
                    ),
                    "load_strategy": table_config.get(  # Default to insert_replace
                        "load_strategy", EC.LS_INSERT_REPLACE
                    ),
                    "target_table_name": operator_config["tables"][table].get(
                        "target_table_name", table
                    ),
                    "target_schema_name": target_schema_name,
                    "target_schema_suffix": target_schema_suffix,
                    "target_database_name": target_database_name,
                }
            )
            # Atomic DAG only works with full refresh and subsequent strategies!
            assert table_config["extract_strategy"] in (
                EC.ES_FULL_REFRESH,
                EC.ES_SUBSEQUENT,
            )
            table_task = el_operator(**table_config)
            kickoff >> table_task >> final

    return (dag,)



from ewah.hooks.sql_base import EWAHSQLBaseHook
from ewah.constants import EWAHConstants as EC

from google.cloud import bigquery
from tempfile import NamedTemporaryFile
from datetime import timedelta

from typing import List, Dict, Union, Type, Optional

import os


class EWAHBigQueryHook(EWAHSQLBaseHook):

    _ATTR_RELABEL: dict = {}

    conn_name_attr = "ewah_bigquery_conn_id"
    default_conn_name = "ewah_bigquery_default"
    conn_type = "ewah_bigquery"
    hook_name = "EWAH BigQuery Connection"

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        return {
            "hidden_fields": ["extra", "password", "login", "schema", "host", "port"],
            "relabeling": {},
        }

    @staticmethod
    def get_connection_form_widgets() -> dict:
        """Returns connection widgets to add to connection form"""
        # from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from ewah.ewah_utils.widgets import EWAHTextAreaWidget
        from wtforms import StringField

        return {
            "extra__ewah_bigquery__service_account_json": StringField(
                "Service Account JSON",
                widget=EWAHTextAreaWidget(rows=12),
            ),
        }

    class bq_cursor(object):
        """Wrapper around the BigQuery client to work similarly to other db cursor.

        :param client: An initialized and authenticated bigquery Client object.
        :param return_dict: Set to true to return lists of dictionaries in fetach calls,
            otherwise fetches return lists of lists. Defaults to True.
        """

        def __init__(
            self,
            outer_class,
            client: Type[bigquery.client.Client],
            return_dict: bool = True,
        ):
            self.client = client
            self.outer = outer_class  # access outer class
            self.return_dict = return_dict
            self.latest_query = None

        def execute(self, sql, params=None, commit=True):
            if not commit:
                self.outer.log.info("Warning: BigQuery always auto-commits!")

            # params have to be supplied via a job_config object!
            # only scalar implemented - array and structs are TODO!
            if params:
                type_mapping = EC.QBC_TYPE_MAPPING.get(EC.DWH_ENGINE_BIGQUERY)
                _default = type_mapping[EC.QBC_TYPE_MAPPING_DEFAULT]
                job_config = bigquery.QueryJobConfig(
                    query_parameters=[
                        bigquery.ScalarQueryParameter(
                            key,
                            type_mapping.get(type(value), _default),
                            value,
                        )
                        for key, value in params.items()
                    ]
                )
            else:
                job_config = None

            self.latest_query = self.client.query(sql, job_config=job_config)

        def fetchall(self):
            return self.fetch_all()

        def fetch_all(self):
            if not self.latest_query:
                return []

            dict_result = [dict(row) for row in latest_query.result()]
            if self.return_dict:
                return dict_result

            keys = dict_result[0].keys()
            return [[row[key] for key in keys] for row in dict_result]

        def fetch_in_batches(self, batch_size: int):
            """Generator to iterate over pages of a custom size."""
            if not self.latest_query:
                return []
            for page in self.latest_query.result(page_size=batch_size).pages:
                self.outer.log.info("Yielding a page...")
                dict_result = [dict(row) for row in page]
                if self.return_dict:
                    yield dict_result
                else:
                    keys = dict_result[0].keys()
                    yield [[row[key] for key in keys] for row in dict_result]

    def _get_db_conn(self):
        # Authenticate via service account json stored in a temporary file
        with NamedTemporaryFile() as cred_json:
            cred_json.seek(0)
            cred_json.write(self.conn.service_account_json.encode())
            cred_json.seek(0)
            cred_json.flush()
            conn = bigquery.client.Client.from_service_account_json(
                os.path.abspath(cred_json.name)
            )
        return conn

    def _get_cursor(self):
        return self.bq_cursor(outer_class=self, client=self.dbconn, return_dict=False)

    def _get_dictcursor(self):
        return self.bq_cursor(outer_class=self, client=self.dbconn, return_dict=True)

    def execute(
        self, sql: str, params: Optional[dict] = None, commit: bool = False, cursor=None
    ) -> None:
        self.log.info(
            "Executing SQL:\n\n{0}\n\nWith params:\n{1}".format(
                sql,
                "\n".join(
                    [
                        "{0}: {1}".format(key, str(value))
                        for (key, value) in params.items()
                    ]
                )
                if params
                else "No params!",
            )
        )
        (cursor or self.cursor).execute(sql=sql.strip(), params=params, commit=commit)

    def get_data_from_sql(
        self, sql: str, params: Optional[dict] = None, return_dict: bool = True
    ) -> Union[List[list], List[dict]]:
        cur = self.dictcursor if return_dict else self.cursor
        self.execute(sql, params=params, cursor=cur, commit=False)
        return cur.fetchall()

    def get_data_in_batches(
        self,
        sql: str,
        params: Optional[dict] = None,
        return_dict: bool = True,
        batch_size: int = 25000,
    ):
        cur = self.dictcursor if return_dict else self.cursor
        self.execute(sql, params=params, cursor=cur)
        data = []
        for batch in cur.fetch_in_batches(batch_size=batch_size):
            data += batch
            if len(data) >= batch_size:
                yield data
                data = []
        if data:
            yield data



from ewah.hooks import hook_class_names


def get_provider_info():
    return {
        "package-name": "ewah",
        "name": "ewah",
        "description": "",
        "hook-class-names": hook_class_names,
        "versions": ["1.0.0"],
    }



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC
from ewah.hooks.pipedrive import EWAHPipedriveHook

from typing import Optional


class EWAHPipedriveOperator(EWAHBaseOperator):

    _NAMES = ["pd", "pipedrive"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
    }

    _CONN_TYPE = EWAHPipedriveHook.conn_type

    def __init__(
        self,
        pipedrive_object: Optional[str] = None,
        additinal_params: Optional[dict] = None,
        *args,
        **kwargs
    ) -> None:
        self.pipedrive_object = pipedrive_object or kwargs.get("target_table_name")
        self.additinal_params = additinal_params or {}
        super().__init__(*args, **kwargs)

    def ewah_execute(self, context: dict) -> None:
        self.log.info("Fetching data from Pipedrive for {self.pipedrive_object}...")
        for batch in self.source_hook.get_data_in_batches(
            self.pipedrive_object, **self.additinal_params
        ):
            self.upload_data(batch)



from ewah.constants import EWAHConstants as EC
from ewah.operators.base import EWAHBaseOperator

import boto3


class EWAHDynamoDBOperator(EWAHBaseOperator):

    _NAMES = ["dynamodb"]

    # For incremental loading, use Kinesis Firehose to push changes to S3
    # and use S3 operator instead
    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
    }

    _REQUIRES_COLUMNS_DEFINITION = False

    def __init__(
        self,
        source_table_name=None,  # defaults to target_table_name
        pagination_limit=None,  # optionally set a pagination limit
        region_name=None,  # must provide region, alternatively via connection
        filter_expression=None,
        *args,
        **kwargs
    ):

        source_table_name = source_table_name or kwargs.get("target_table_name")
        _msg = "DynamoDBOperator requires primary_key_column_name!"
        assert kwargs.get("primary_key_column_name"), _msg
        super().__init__(*args, **kwargs)
        self.source_table_name = source_table_name
        self.pagination_limit = pagination_limit
        self.region_name = region_name
        self.filter_expression = filter_expression

    def ewah_execute(self, context):
        for batch in self.source_hook.get_dynamodb_data_in_batches(
            table_name=self.source_table_name,
            region=self.region_name,
            pagination_limit=self.pagination_limit,
            filter_expression=self.filter_expression,
        ):
            self.upload_data(batch)



from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC

from ewah.hooks.base import EWAHBaseHook as BaseHook

from sshtunnel import SSHTunnelForwarder
from tempfile import NamedTemporaryFile
from datetime import timedelta
from pytz import timezone

import os

from typing import Optional, List, Dict, Any, Union


class EWAHSQLBaseOperator(EWAHBaseOperator):
    # Don't overwrite the _NAMES list to avoid accidentally exposing the SQL
    # base operator
    # _NAMES = []

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,
        EC.ES_SUBSEQUENT: True,
    }

    def __init__(
        self,
        source_schema_name: Optional[str] = None,
        source_table_name: Optional[
            str
        ] = None,  # defaults to same as target_table_name
        source_database_name: Optional[str] = None,  # bigquery: project id
        sql_select_statement: Optional[str] = None,  # Alternative to specifying table
        timestamp_column: Optional[str] = None,
        where_clauses: Optional[Union[str, List[str]]] = None,
        extra_params: Optional[dict] = None,
        batch_size: int = 100000,
        use_limits_for_batches: bool = False,
        *args,
        **kwargs
    ):
        source_table_name = source_table_name or kwargs["target_table_name"]

        # default subsequent_field to timestamp_column or primary_key_column_name
        if kwargs.get("extract_strategy") == EC.ES_SUBSEQUENT:
            kwargs["subsequent_field"] = kwargs.get(
                "subsequent_field",
                timestamp_column or kwargs.get("primary_key_column_name", None),
            )

        super().__init__(*args, **kwargs)

        if isinstance(where_clauses, str):
            where_clauses = [where_clauses]

        if self.extract_strategy == EC.ES_INCREMENTAL:
            assert timestamp_column, "Incremental loading must have timestamp column!"

        if not sql_select_statement:
            assert source_schema_name
            assert source_table_name
            if self.columns_definition:
                columns_sql = "\n\t  {0}{1}{0}".format(
                    self._SQL_COLUMN_QUOTE,
                    "{0}\n\t, {0}".format(self._SQL_COLUMN_QUOTE).join(
                        self.columns_definition.keys()
                    ),
                )
            else:
                columns_sql = "*"
            sql_select_statement = self._SQL_BASE.format(
                columns=columns_sql,
                schema=source_schema_name,
                table=source_table_name,
                database=source_database_name,
            )

        self.sql = self._SQL_BASE_SELECT.format(select_sql=sql_select_statement)
        self.extra_params = extra_params
        self.timestamp_column = timestamp_column
        self.where_clauses = where_clauses
        self.batch_size = batch_size
        self.use_limits_for_batches = use_limits_for_batches

    def ewah_execute(self, context):
        # called, potentially with a data_from and data_until

        params = self.extra_params or {}
        where_clauses = self.where_clauses or []
        if self.data_from:
            where_clauses.append(
                "{0} >= {1}".format(
                    self.timestamp_column, self._SQL_PARAMS.format("data_from")
                )
            )
            params["data_from"] = self.data_from
        if self.data_until:
            where_clauses.append(
                "{0} <= {1}".format(
                    self.timestamp_column, self._SQL_PARAMS.format("data_until")
                )
            )
            params["data_until"] = self.data_until
        if self.subsequent_field and self.test_if_target_table_exists():
            where_clauses.append(
                "{0} > {1}".format(
                    self.subsequent_field, self._SQL_PARAMS.format("previous_max_value")
                )
            )
            params["previous_max_value"] = self.get_max_value_of_column(
                self.subsequent_field
            )

        where_clauses = where_clauses or ["1 = 1"]
        sql = self.sql.format("\n  AND ".join(where_clauses))
        for batch in self.source_hook.get_data_in_batches(
            sql=sql,
            params=params or None,  # Don't supply empty dict as params!
            return_dict=True,
            batch_size=self.batch_size,
            use_limits=self.use_limits_for_batches,
            order_by_columns=self.update_on_columns,
        ):
            self.upload_data(batch)



from ewah.operators.sql_base import EWAHSQLBaseOperator
from ewah.hooks.postgres import EWAHPostgresHook


class EWAHPostgresOperator(EWAHSQLBaseOperator):

    _NAMES = ["pgsql", "postgres", "postgresql"]

    _SQL_BASE = 'SELECT\n{columns}\nFROM "{schema}"."{table}"\n'
    _SQL_BASE_SELECT = (
        "WITH raw_data AS (\n\n{select_sql}\n\n) SELECT * FROM raw_data\nWHERE {{0}}"
    )
    _SQL_COLUMN_QUOTE = '"'
    _SQL_PARAMS = "%({0})s"

    _CONN_TYPE = EWAHPostgresHook.conn_type



from ewah.uploaders.base import EWAHBaseUploader
from ewah.constants import EWAHConstants as EC

from tempfile import NamedTemporaryFile
from datetime import datetime
from decimal import Decimal
import gspread
import json
import os


def monkeypatch_values_update(func_to_call):
    def monkeypatch_func(range_name, params, body):
        values = body.get("values")
        if values:
            values = [
                [
                    float(v)
                    if isinstance(v, Decimal)
                    else (
                        v.strftime("%Y-%m-%d %H:%M:%S%z")
                        if isinstance(v, datetime)
                        else v
                    )
                    for v in outer
                ]
                for outer in values
            ]
            body["values"] = values
        return func_to_call(range_name, params=params, body=body)

    return monkeypatch_func


class EWAHGSheetsUploader(EWAHBaseUploader):
    """
    The Google Sheets Hook works a little bit different to other hooks.
        - it can only have 1 upload call per task execution
        - it deletes all data from the sheet and re-uploads it to the sheet
        - the target schema name is the workbook key
        - the target table name is the worksheet name
    """

    def __init__(self, *args, **kwargs):
        self._upload_call_count = 0  # Make sure there is only a single upload
        super().__init__(EC.DWH_ENGINE_GS, *args, **kwargs)

    def detect_and_apply_schema_changes(self, *args, **kwargs):
        # need to overwrite parent function, does nothing
        raise Exception("This function is not implemented for this DWHook!")

    def commit(self):
        self.log.info("Everything is auto-committed")

    def rollback(self):
        raise Exception("Cannot rollback Google Sheets target!")

    def close(self):
        self.log.info("Nothing to close")

    def _create_or_update_table(
        self,
        data,
        table_name,
        schema_name,
        schema_suffix,  # unused but always given
        columns_definition,  # unused but always given
        columns_partial_query,  # unused but always given
        update_on_columns,  # unused but always given
        load_strategy,  # Must be LS_INSERT_REPLACE
        upload_call_count,  # Must be 1
        pk_columns=None,  # must accept arg, but it must also always be []
    ):
        # Google Sheets only works with drop & replace in one go
        assert load_strategy == EC.LS_INSERT_REPLACE
        assert upload_call_count == 1

        if pk_columns:
            raise Exception("Arg pk_columns invalidly supplied!")

        def colnum_string(n):
            # adapted from https://stackoverflow.com/questions/23861680/convert-spreadsheet-number-to-column-letter
            string = ""
            while n > 0:
                n, remainder = divmod(n - 1, 26)
                string = chr(65 + remainder) + string
            return string

        if not drop_and_replace:
            raise Exception("Google Sheets DWHs can only be drop_and_replace!")
        self._upload_call_count += 1
        if not self._upload_call_count == 1:
            raise Exception("Chunking is not possible for Google Sheets DWH!")
        self.log.info("Replacing data in Google Sheets!")

        if not data:
            self.log.info("Nothing to upload!")
            return

        # authorize and get correct worksheet
        servie_acc_file = NamedTemporaryFile()
        self.log.info("Authenticating...")
        credentials = self.dwh_conn.extra_dejson
        credentials = credentials.get("client_secrets", credentials)
        service_acc = json.dumps(credentials)
        filename = os.path.abspath(servie_acc_file.name)
        with open(filename, "w") as f:
            f.write(service_acc)
        gclient = gspread.service_account(filename=filename)

        workbook = gclient.open_by_key(schema_name)
        worksheet = workbook.worksheet(table_name)
        values = worksheet.get_all_values()

        # Delete old data, if any existed
        if values:
            self.log.info("Deleting old data...")
            rows = len(values)
            columns = len(values[0])
            values = [columns * [""]] * rows
            range_notation = "A1:" + colnum_string(columns) + str(rows)
            worksheet.update(range_notation, values)  # delete!

        del values  # free up memory

        # insert new data - need to change the format! use columns definition
        self.log.info("Preparing data for Google Sheets upload...")
        column_header = list(columns_definition.keys())
        upload_data = [column_header]
        while data:
            current_data = data.pop(0)
            upload_data += [[current_data.get(col, "") for col in column_header]]
        range_notation = "A1:" + colnum_string(len(column_header))
        range_notation += str(len(upload_data))
        self.log.info("Uploading data now!")
        worksheet.spreadsheet.values_update = monkeypatch_values_update(
            worksheet.spreadsheet.values_update
        )
        worksheet.update(range_notation, upload_data)
        self.log.info("Upload done.")



from ewah.hooks.sql_base import EWAHSQLBaseHook

from psycopg2 import connect as pg_connect
from psycopg2.extras import RealDictCursor
from typing import Optional, List, Union, Dict, Any


class EWAHPostgresHook(EWAHSQLBaseHook):

    _DEFAULT_PORT = 5432

    _ATTR_RELABEL: dict = {
        "database": "schema",
        "hostname": "host",
        "user": "login",
    }

    conn_name_attr = "ewah_postgres_conn_id"
    default_conn_name = "ewah_postgres_default"
    conn_type = "ewah_postgres"
    hook_name = "EWAH PostgreSQL Connection"

    _LIMIT_SQL = """
        SELECT * FROM ({sql_query}) t
        ORDER BY {order_by_columns}
        LIMIT {limit}
        OFFSET {offset}
    """

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        return {
            "hidden_fields": ["extra"],
            "relabeling": {
                "password": "Password",
                "login": "User",
                "schema": "Database",
                "host": "Hostname / IP",
                "port": "Port (default: 5432)",
            },
        }

    @staticmethod
    def get_connection_form_widgets() -> dict:
        """Returns connection widgets to add to connection form"""
        from flask_appbuilder.fieldwidgets import BS3TextFieldWidget
        from wtforms import StringField

        return {
            f"extra__ewah_postgres__ssh_conn_id": StringField(
                "SSH Connection ID (optional)",
                widget=BS3TextFieldWidget(),
            ),
        }

    def _get_db_conn(self):
        return pg_connect(
            "dbname='{0}' user='{1}' host='{2}' password='{3}' port='{4}'".format(
                self.conn.database,
                self.conn.user,
                self.local_bind_address[0],
                self.conn.password,
                self.local_bind_address[1],
            )
        )

    def _get_cursor(self):
        return self.dbconn.cursor()

    def _get_dictcursor(self):
        return self.dbconn.cursor(cursor_factory=RealDictCursor)

    def execute(
        self, sql: str, params: Optional[dict] = None, commit: bool = False, cursor=None
    ) -> None:
        self.log.info(
            "Executing SQL:\n\n{0}\n\nWith params:\n{1}".format(
                sql,
                "\n".join(
                    [
                        "{0}: {1}".format(key, str(value))
                        for (key, value) in params.items()
                    ]
                )
                if params
                else "No params!",
            )
        )
        (cursor or self.cursor).execute(sql.strip(), vars=params)
        if commit:
            self.commit()

    def get_data_from_sql(
        self, sql: str, params: Optional[dict] = None, return_dict: bool = True
    ) -> Union[List[list], List[dict]]:
        cur = self.dictcursor if return_dict else self.cursor
        self.execute(sql, params=params, cursor=cur, commit=False)
        return cur.fetchall()

    def get_data_in_batches(
        self,
        sql: str,
        params: Optional[dict] = None,
        return_dict: bool = True,
        batch_size: int = 100000,
    ):
        cur = self.dictcursor if return_dict else self.cursor
        self.execute(sql, params=params, cursor=cur, commit=False)
        while True:
            data = cur.fetchmany(batch_size)
            if data:
                yield data
            else:
                break
